# pylint: skip-file
from ioc import require
from ioc import provide
from ioc import inject

from . import services
from .asgi import get_asgi_application
from .auth import *
from .decorators import *
from .dependency import ContextualInjected
from .dependency import Injected
from .routers import FastAPIRouter
from .viewsets import GenericViewSet
from .views import AnonymousResource
from .views import Resource


__all__ = [
    'action',
    'get_asgi_application',
    'inject',
    'provide',
    'response_model',
    'BearerClaims',
    'BearerSubjectResolver',
    'BearerTokenDecoder',
    'CurrentSubject',
    'Resource',
    'RequestClaimSet',
]



"""Provides the functions to create the ASGI application."""
import inspect
import logging
import warnings

import fastapi
import unimatrix.runtime
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import JSONResponse
from ioc.exc import UnsatisfiedDependency
from starlette.middleware.trustedhost import TrustedHostMiddleware
from unimatrix.conf import settings
from unimatrix.ext.model.exc import CanonicalException
from unimatrix.ext.model.exc import FeatureNotSupported

from .resources import HealthCheckResource
from .resources import ServiceMetadataResource
from .routers import FastAPIRouter


logger = logging.getLogger('endpoints')
DEFAULT_ROUTER = FastAPIRouter()
DEFAULT_SECRET_KEY = '0' * 64


def _ensure_secret_key():
    try:
        settings.SECRET_KEY
    except AttributeError:
        warnings.warn(
            "CAUTION: The SECRET_KEY setting is not specified, using the "
            "default of {DEFAULT_SECRET_KEY}."
        )
        os.environ['SECRET_KEY'] = DEFAULT_SECRET_KEY


class FastAPI(fastapi.FastAPI):
    """A :class:`fastapi.FastAPI` subclass that provides integration with
    the Unimatrix Framework.
    """
    endpoints = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.endpoints is None:
            self.endpoints = DEFAULT_ROUTER

    def add_resource_router(self, router):
        """Add a :class:`~frf.FastAPIRouter` to the ASGI application."""
        router.add_to_app(self)

    def add_resource(self, cls, base_path=None):
        """Add a RESTFUL API resource `cls` to the default router."""
        app = self
        if base_path is None:
            base_path = f'/{cls.get_resource_name()}'
        if not base_path:
            base_path = '/'
        if not str.startswith(base_path, '/'):
            base_path = f'/{base_path}'

        # Iterate over all members of the class and find actions.
        for _, func in inspect.getmembers(cls):
            if not hasattr(func, 'action'):
                continue
            func.action.add_to_app(app, base_path, cls)

        if hasattr(cls, 'create'):
            cls.as_method_handler(
                app, f'{base_path}', 'POST', 'create')
        if hasattr(cls, 'retrieve'):
            cls.as_method_handler(
                app, f'{base_path}/{{resource_id}}', 'GET', 'retrieve')
        if hasattr(cls, 'list'):
            cls.as_method_handler(
                app, f'{base_path}', 'GET', 'list')
        if hasattr(cls, 'update'):
            cls.as_method_handler(
                app, f'{base_path}/{{resource_id}}', 'PATCH', 'update')
        if hasattr(cls, 'delete'):
            cls.as_method_handler(
                app, f'{base_path}/{{resource_id}}', 'DELETE', 'delete')
        if hasattr(cls, 'replace'):
            cls.as_method_handler(
                app, f'{base_path}/{{resource_id}}', 'PUT', 'replace')

        self.include_router(cls.router,
            tags=[cls.resource_name or cls.__name__])

        cls.asgi = self


async def _handle_canonical_exception(request, exc):
    if isinstance(exc, UnsatisfiedDependency):
        return await _handle_canonical_exception(request, FeatureNotSupported())
    if isinstance(exc, CanonicalException):
        if exc.http_status_code >= 500:
            exc.log(logger.exception)
        return JSONResponse(
            status_code=exc.http_status_code,
            content=exc.as_dict()
        )
    else:
        raise NotImplementedError


def get_asgi_application(allowed_hosts=None, **kwargs):
    """Return a :class:`fastapi.FastAPI` instance. This is the main ASGI
    application.

    The :func:`get_asgi_application()` gathers configuration from various
    sources in order to configure a basic :class:`fastapi.FastAPI` instance. The
    order is as defined below:

    - :attr:`unimatrix.conf.settings` - this module loads all uppercased
      variables in the Python module (or package) indicated by the environment
      variable ``UNIMATRIX_SETTINGS_MODULE`` (see below).
    - :mod:`unimatrix.environ`

    All common environment variables specified by the Unimatrix Framework are
    detected by :func:`get_asgi_application()`. Additional configuration may
    be provided by setting the environment variable
    ``UNIMATRIX_SETTINGS_MODULE``, holding the qualified name of a Python
    module. See also :ref:`index-environment-variables`.
    """
    app = FastAPI(
        openapi_url=getattr(settings, 'OPENAPI_URL', '/openapi.json'),
        docs_url=getattr(settings, 'DOCS_URL', '/'),
        exception_handlers={
            CanonicalException: _handle_canonical_exception,
            UnsatisfiedDependency: _handle_canonical_exception
        }
    )
    _ensure_secret_key()
    if settings.HTTP_CORS_ENABLED:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=settings.HTTP_CORS_ALLOW_ORIGINS,
            allow_credentials=settings.HTTP_CORS_ALLOW_CREDENTIALS,
            allow_methods=settings.HTTP_CORS_ALLOW_METHODS,
            allow_headers=settings.HTTP_CORS_ALLOW_HEADERS,
            expose_headers=settings.HTTP_CORS_EXPOSE_HEADERS,
            max_age=settings.HTTP_CORS_TTL
        )

    # Always enable the TrustedHostMiddleware.
    app.add_middleware(
        TrustedHostMiddleware,
        allowed_hosts=(
            allowed_hosts or getattr(settings, 'HTTP_ALLOWED_HOSTS', [])
        )
    )

    app.add_resource(HealthCheckResource, base_path='.well-known/health')
    app.add_resource(ServiceMetadataResource, base_path='.well-known/self')

    @app.on_event('startup')
    async def on_startup():
        await unimatrix.runtime.on('boot')

    @app.on_event('shutdown')
    async def on_shutdown():
        await unimatrix.runtime.on('shutdown')

    return app



# pylint: skip-file
import ioc

import frf
from frf.lib.test import ProtectedResourceTestCase
from .base import ProtectedResource


class MinimalBearerTokenTestCase(ProtectedResourceTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, ProtectedResource)
    ]

    def get_asgi_application(self):
        return frf.get_asgi_application(allowed_hosts=['*'])

    def test_list_returns_200(self):
        response = self.request(self.client.get, self.base_path)
        self.assertEqual(response.status_code, 200, response.json())



"""Declares low-level functions and classes for request handling."""
import functools
import inspect
import logging
import re
from collections import OrderedDict
from inspect import Parameter

from fastapi import APIRouter
from fastapi import Path
from fastapi import Request
from fastapi import Security
from fastapi.security import HTTPBearer
from fastapi.security import HTTPAuthorizationCredentials

from . import exceptions
from . import auth


class ResourceMetaclass(type):
    """Constructs a :class:`APIView` class."""

    def __new__(cls, name, bases, attrs):
        super_new = super().__new__
        if name in ('APIView', 'Resource'):
            return super_new(cls, name, bases, attrs)
        attrs['router'] = APIRouter()
        return super_new(cls, name, bases, attrs)


class Resource(metaclass=ResourceMetaclass):
    """The base class for all API views (request handlers)."""
    authentication_classes = [None]

    #: The required audience for bearer tokens.
    audience = []

    #: The HTTP security scheme.
    security = HTTPBearer(auto_error=False, scheme_name='Bearer')

    #: Indicates if cookie authentication is allowed for this endpoint.
    allow_cookie_auth = False

    #: The list of authentication classes that are used.

    #: The name of the cookie holding the Bearer token.
    auth_cookie_name = 'sessionid' # nosec

    #: The list of issuers that this resource accepts.
    issuers = []

    #: The default logger for endpoints
    logger = logging.getLogger('endpoints')

    #: The human-readable resource name.
    resource_name = None

    #: The list of scopes that are required to access this resource using a
    #: Bearer token.
    required_scopes = []

    #: The list of allowed HTTP methods.
    http_method_names = ['get', 'post', 'put', 'patch', 'delete', 'head',
        'options', 'trace']

    @classmethod
    def get_resource_name(cls):
        """Return the name that will be used in constructing the
        URL paths for this :class:`Resource`.
        """
        # Some naieve pluralizing
        name = str.lower(cls.__name__)
        if not str.endswith(name, 's'):
            name = name + 's'
        return re.sub('(Resource|View|ViewSet)$', '', name)

    @classmethod
    def is_protected(cls):
        """Return a boolean indicating if the resource is protected."""
        return bool(cls.authentication_classes)

    @classmethod
    def as_method_handler(cls, app, path, method, handler, **initkwargs):
        """Main entry point for a request-response process for the given request
        method.
        """
        method = str.lower(method)
        if not hasattr(app, method) or not hasattr(cls, handler):
            raise ValueError(f"Unsupported request method: {method}")
        for key in initkwargs:
            if key in cls.http_method_names:
                raise TypeError(
                    'The method name %s is not accepted as a keyword argument '
                    'to %s().' % (key, cls.__name__)
                )
            if not hasattr(cls, key):
                raise TypeError("%s() received an invalid keyword %r. as_view "
                                "only accepts arguments that are already "
                                "attributes of the class." % (cls.__name__, key))


        # Instantiate the view class so we can use the signature of the
        # request handler, which is used by FastAPI to generated the API
        # interface.
        view = cls(**initkwargs)
        register_endpoint = getattr(cls.router, method)
        f = getattr(view, handler)
        response_model = getattr(f, 'response_model', None)

        # Do some black magic here to trick FastAPI to inject the request
        # and response.
        signature = inspect.signature(f)
        call_args = OrderedDict()

        # Add the Request object as the first argument if it is not provided.
        inject_request = 'request' in signature.parameters
        inject_id = handler in ['retrieve', 'update', 'delete', 'replace']
        call_args['request'] = inspect.Parameter(
            'request',
            Parameter.POSITIONAL_OR_KEYWORD,
            annotation=Request
        )

        # Update the call_args dictionary with the remaining parameters
        # and create a new call signature. Do that before the other arguments
        # to prevent argument order errors.
        call_args.update({x: y for x, y in list(signature.parameters.items())
            if x not in call_args})

        if inject_id:
            p = signature.parameters.get('resource_id')
            call_args['resource_id'] = inspect.Parameter(
                'resource_id',
                Parameter.POSITIONAL_OR_KEYWORD,
                default=Path(...,
                    title="Identifies the {cls.get_resource_name()} object."
                ),
                annotation=p.annotation\
                    if (p and p.annotation != inspect._empty) else str
            )

        # Ensure that the Authorization header is requested if any
        # authentication classes are defined.
        if cls.authentication_classes:
            call_args['__authorization'] = inspect.Parameter(
                '__authorization',
                Parameter.POSITIONAL_OR_KEYWORD,
                default=Security(cls.security),
                annotation=HTTPAuthorizationCredentials
            )
            call_args['__claims'] = inspect.Parameter(
                '__claims',
                Parameter.POSITIONAL_OR_KEYWORD,
                default=auth.RequestClaims(
                    allow_cookies=cls.allow_cookie_auth,
                    cookie_name=cls.auth_cookie_name
                )
            )

        # Create the final signature
        signature = signature.replace(
            parameters=list(call_args.values()))

        # Some ugly method but this whole function has to be rewritten
        # anyway.
        endpoint_name = re.sub(
            '(Resource|View|ViewSet)$', '',
            getattr(cls, 'endpoint_name', cls.__name__)
        )

        endpoint_params = {
            'response_model': response_model,
            'name': str.lower(f'{endpoint_name}.{handler}')
        }

        @register_endpoint(path, **endpoint_params)
        @functools.wraps(f)
        async def request_handler(request: Request, **kwargs):
            view = cls(**initkwargs)
            view.cls = cls
            view.initkwargs = initkwargs
            view.funckwargs = dict({x: y for x,y in dict.items(kwargs)
                if not str.startswith(x, '__')})
            if inject_request:
                view.funckwargs['request'] = request
            return await view.dispatch(getattr(view, handler), request, **kwargs)

        request_handler.__signature__ = signature
        request_handler.__annotations__ = call_args

        return request_handler

    async def dispatch(self, handler, request, *args, **kwargs):
        """`.dispatch()` is pretty much the same as Django's regular dispatch,
        but with extra hooks for startup, finalize, and exception handling.
        """
        self.args = args
        self.kwargs = kwargs
        try:
            request = await self.initialize_request(request, *args, **kwargs)
            self.request = request
            await self.initial(request, *args, **kwargs)
            response = await handler(**dict(self.funckwargs))
        except Exception as exc: # pylint: disable=broad-except
            response = await self.handle_exception(exc)

        self.response = await self.finalize_response(request, response)
        return response

    async def finalize_response(self, request, response):
        """Returns the final response object."""
        return response

    async def initial(self, request, *args, **kwargs):
        """Runs anything that needs to occur prior to calling the method
        handler.
        """
        self.claims = claims = kwargs.pop('__claims', None)
        if self.is_protected():
            await self.verify_claims(claims)

    async def handle_exception(self, exc):
        """Handle any exception that occurs, by returning an appropriate
        response, or re-raising the error.
        """
        self.raise_uncaught_exception(exc)

    async def http_method_not_allowed(self, request, *args, **kwargs):
        """If `request.method` does not correspond to a handler method,
        determine what kind of exception to raise.
        """
        raise exceptions.MethodNotAllowed(request.method)

    async def initialize_request(self, request, *args, **kwargs):
        """Returns the initial request object."""
        return request

    async def verify_claims(self, claims):
        """Verify the claims specified by the Bearer token."""
        claims.verify_expired()
        if self.audience:
            claims.verify_audience(self.audience)
        if self.required_scopes:
            claims.verify_scopes(self.required_scopes)
        if self.issuers:
            claims.verify_issuer(self.issuers)

    def raise_uncaught_exception(self, exc):
        """Raise an exception that is not caught by a handler."""
        raise exc


class APIView(Resource):
    pass


class AnonymousResource(Resource):
    """Like :class:`Resource`, but does not authenticate the request using
    a Bearer token.
    """
    authentication_classes = []



# pylint: skip-file
import frf
from frf.lib.test import ResourceEndpointTestCase


class ProtectedResource(frf.Resource):

    async def list(self):
        return {}


class ProtectedResourceTestCase(ResourceEndpointTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, ProtectedResource)
    ]

    def get_asgi_application(self):
        return frf.get_asgi_application(allowed_hosts=['*'])

    def test_list_returns_403(self):
        response = self.client.get(self.base_path)
        self.assertEqual(response.status_code, 401, response.json())

    def test_list_returns_bearer_token_missing(self):
        response = self.client.get(self.base_path)
        dto = response.json()
        self.assertEqual(response.status_code, 401)
        self.assertIn('code', dto, dto)
        self.assertEqual(dto['code'], 'BEARER_TOKEN_MISSING')



"""Declares :class:`HealthCheckService`."""


class HealthCheckService:
    """Exposes an API to perform health checks. The
    :class:`HealthCheckService` makes the distinction between two types of
    health checks.

    A **readyness** check indicates if the application is ready to serve
    traffic. If a readyness check fails, the deployment infrastructure may
    choose to stop sending traffic to the application.

    A **liveness** check indicates if the application is alive or dead. A
    failing liveness check may result in the application process being
    killed.

    :class:`HealthCheckService` is injected with the name ``HealthCheckService``
    using the :mod:`ioc` module. To customize the behavior of the health checks,
    a subclass may be provided by injecting it at application startup:

    .. code:: python

        import frf

        from myproject import MyHealthCheckService


        app = frf.get_asgi_application()

        @app.on_event('startup')
        async def boot():
            frf.provide('HealthCheckService',
                MyHealthCheckService(), force=True)
    """

    async def is_live(self) -> bool:
        """Return a boolean indicating if the application is live."""
        return True

    async def is_ready(self) -> bool:
        """Return a boolean indicating if the application is ready."""
        return True



"""Declares various functions to inject dependencies."""
from typing import Any

import ioc
from fastapi import Depends


def Injected(name, *args, **kwargs):
    """Like :func:`fastapi.Depends`, but takes the identifier of a dependency
    that is injected through :func:`frf.provide`.
    """
    return Depends(lambda: ioc.require(name), *args, **kwargs)


def ContextualInjected(dependency: str, use_cache: bool = True) -> Any:
    """Like :func:`Injected`, but sets up a context."""
    async def context_factory():
        provider = ioc.require(dependency)
        async with provider.as_context() as ctx:
            yield ctx
    return Depends(context_factory, use_cache=use_cache)



"""Declares :class:`HTTPStatusResource`."""
from fastapi import Response

from ..decorators import action
from ..dependency import Injected
from ..views import AnonymousResource


class HealthCheckResource(AnonymousResource):
    """Exposes endpoints for live- and readyness checks."""
    __module__ = 'frf.resources'
    resource_name = 'Health'

    @action(name='live', detail=False, methods=['get'])
    async def live(self,
        response: Response,
        check=Injected('HealthCheckService')
    ):
        """Perform a liveness check."""
        try:
            if not await check.is_live():
                response.status_code = 503
        except Exception: # pylint: disable=broad-except
            response.status_code = 503

    @action(name='ready', detail=False, methods=['get'])
    async def ready(self,
        response: Response,
        check=Injected('HealthCheckService')
    ):
        """Perform a readiness check."""
        try:
            if not await check.is_ready():
                response.status_code = 503
        except Exception: # pylint: disable=broad-except
            response.status_code = 503



# pylint: skip-file
import frf
from frf.lib.test import ResourceEndpointTestCase


class AnonymousResource(frf.AnonymousResource):

    async def list(self):
        return {}


class AnonymousResourceTestCase(ResourceEndpointTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, AnonymousResource)
    ]

    def get_asgi_application(self):
        return frf.get_asgi_application(allowed_hosts=['*'])

    def test_list_returns_200(self):
        response = self.client.get(self.base_path)
        self.assertEqual(response.status_code, 200, response.json())



import ioc
import uvicorn

import frf
from frf import AnonymousResource
from frf import get_asgi_application


class TemplateService:

    async def render_to_string(self, template_name, ctx):
        return template_name


class SinglePageApplication(AnonymousResource):

    @frf.spa('dashboard.html.j2')
    async def dashboard(self):
        return {}


app = get_asgi_application(allowed_hosts=['*'])
app.add_resource(SinglePageApplication)


if __name__ == '__main__':
    ioc.provide('TemplateService', TemplateService())
    ioc.provide('APIMetadataService', frf.services.APIMetadataService())
    uvicorn.run(app, host="127.0.0.1", port=8000)



# pylint: skip-file
#
# Do not put more than one class in a file, this is for example purposes
# only.
import os
from typing import Optional

import ioc
import frf
import uvicorn
from pydantic import BaseModel
from unimatrix.ext.model import Repository


class Subject(BaseModel):
    username: str
    email: Optional[str] = None
    full_name: Optional[str] = None
    disabled: Optional[bool] = None


class SubjectRepository(Repository):
    subjects = {
        "johndoe": {
            "username": "johndoe",
            "full_name": "John Doe",
            "email": "johndoe@example.com",
            "secret": "secret_for_john",
            "disabled": False,
        },
        "alice": {
            "username": "alice",
            "full_name": "Alice Wonderson",
            "email": "alice@example.com",
            "secret": "secret_for_alice",
            "disabled": True,
        },
    }

    async def get(self, username):
        return Subject(**self.subjects[username])


class BearerSubjectResolver(frf.BearerSubjectResolver):

    @frf.inject('subjects', 'SubjectRepository')
    async def get_subject(self, claims, subjects):
        return await subjects.get(claims['sub'])


class ProtectedResource(frf.Resource):

    async def list(self, subject=frf.CurrentSubject):
        return {
            "message": (
                f"Hello, {subject.full_name}. "
                "You succesfully accessed the protected resource!"
            )
        }


app = frf.get_asgi_application(allowed_hosts=['*'])
app.add_resource(ProtectedResource)

@app.on_event('startup')
async def boot():
    frf.provide('SubjectRepository', SubjectRepository())
    frf.provide('BearerSubjectResolver', BearerSubjectResolver(), force=True)


if __name__ == '__main__':
    uvicorn.run(app, host="127.0.0.1", port=8000)



"""Supplies authentication functions."""
import logging

import ioc.exc
from fastapi import Cookie
from fastapi import Depends
from fastapi import Request
from unimatrix.ext import crypto
from unimatrix.ext import jose

from .exceptions import BearerTokenMissing
from .exceptions import InvalidAuthorizationScheme
from .exceptions import BearerTokenUnverifiable


logger = logging.getLogger('endpoints')


def get_bearer_token(request: Request):
    header = request.headers.get('Authorization')
    value = None
    if header:
        scheme, _, value = str.partition(header, ' ')
        if str.lower(scheme) != 'bearer':
            raise InvalidAuthorizationScheme(scheme, ['Bearer'])
    return value


def Session(name, enabled=False):
    if not enabled:
        return Depends(lambda: None)

    async def f(request: Request, sessionid: str = Cookie(None, alias=name)):
        return request.cookies.get(name)

    return Depends(f)


def RequestClaims(allow_cookies=False, cookie_name='session', required=True):
    session = Session(cookie_name, allow_cookies)
    async def f(request: Request, bearer=Bearer, sessionid:str=session):
        if not bearer and allow_cookies:
            bearer = sessionid
            if bearer:
                logger.debug("Received Bearer token from cookie %s",
                    cookie_name)
        if not bearer and required:
            logger.debug("Bearer token not included in headers or cookies.")
            raise BearerTokenMissing

        if bearer:
            try:
                return await jose.payload(str.encode(bearer))
            except ioc.exc.UnsatisfiedDependency:
                raise
            except (crypto.Signature.InvalidSignature, LookupError):
                logger.debug(
                    "Received Bearer token with invalid signature or missing "
                    "key.")
                if required:
                    raise BearerTokenUnverifiable

    return Depends(f)


Bearer = Depends(get_bearer_token)



# pylint: skip-file
from unittest.mock import AsyncMock

import ioc
from unimatrix.lib.datastructures import DTO

from ..lib.test import ResourceEndpointTestCase
from ..asgi import get_asgi_application
from ..decorators import spa
from ..views import AnonymousResource


class SinglePageApplication(AnonymousResource):

    @spa('foo.html.j2')
    async def dashboard(self):
        return {'foo': 1, 'bar': 2}


class SinglePageApplicationDecoratorTestCase(ResourceEndpointTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, SinglePageApplication)
    ]

    def setUp(self):
        super().setUp()
        self.svc = DTO(
            render_to_string=AsyncMock(return_value="Hello world!")
        )
        ioc.provide('TemplateService', self.svc)

    def get_asgi_application(self):
        return get_asgi_application(allowed_hosts=['*'])

    def test_template_service_is_invoked(self):
        response = self.client.get('test/dashboard')
        self.assertEqual(response.status_code, 200)
        self.assertEqual(response.text, "Hello world!")



# pylint: skip-file
from unittest.mock import AsyncMock

import ioc

import frf
from frf.lib.test import ResourceEndpointTestCase

from .. import HealthCheckResource


class HealthCheckResourceTestCase(ResourceEndpointTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, HealthCheckResource)
    ]
    headers = {}

    def setUp(self):
        super().setUp()
        self.svc = ioc.provide('HealthCheckService',
            frf.services.HealthCheckService(),
            force=True)

    def get_asgi_application(self):
        return frf.get_asgi_application(allowed_hosts=['*'])

    def test_live_returns_200(self):
        response = self.request(self.client.get, '/.well-known/health/live')
        self.assertEqual(response.status_code, 200, response.json())

    def test_live_returns_503_on_failure(self):
        self.svc.is_live = AsyncMock(return_value=False)
        response = self.request(self.client.get, '/.well-known/health/live')
        self.assertEqual(response.status_code, 503, response.json())

    def test_live_returns_503_on_exception(self):
        self.svc.is_live = AsyncMock(side_effect=Exception)
        response = self.request(self.client.get, '/.well-known/health/live')
        self.assertEqual(response.status_code, 503, response.json())

    def test_ready_returns_200(self):
        self.svc.is_ready = AsyncMock(return_value=False)
        response = self.request(self.client.get, '/.well-known/health/ready')
        self.assertEqual(response.status_code, 503, response.json())

    def test_ready_returns_503_on_failure(self):
        self.svc.is_ready = AsyncMock(return_value=False)
        response = self.request(self.client.get, '/.well-known/health/ready')
        self.assertEqual(response.status_code, 503, response.json())

    def test_ready_returns_503_on_exception(self):
        self.svc.is_ready = AsyncMock(side_effect=Exception)
        response = self.request(self.client.get, '/.well-known/health/ready')
        self.assertEqual(response.status_code, 503, response.json())



"""Declares exceptions used by the FastAPI RESt Framework."""
from fastapi import HTTPException
from unimatrix.ext.model.exc import CanonicalException


class MethodNotAllowed(HTTPException):
    """Raises to indicate that an unsupported HTTP method is requested."""
    status_code = 415
    default_detail = 'Method "{method}" not allowed.'
    default_code = 'method_not_allowed'

    def __init__(self, method, detail=None, code=None):
        if detail is None:
            detail = self.default_detail.format(method=method)
        super().__init__(status_code=self.status_code, detail=detail)


class InvalidAuthorizationScheme(CanonicalException):
    code = 'HTTP_AUTHORIZATION_SCHEME_INVALID'
    http_status_code = 403

    def __init__(self, scheme, supported):
        message = "The Authorization header specified an unknown scheme."
        detail = f"The scheme '{scheme}' is not supported by the server."
        hint = (
            "Use one of the following schemes: "
            f"{str.join(', ', supported)}"
        )
        super().__init__(message=message, hint=hint, detail=detail)


class BearerTokenMissing(CanonicalException):
    code = 'BEARER_TOKEN_MISSING'
    http_status_code = 401


class BearerTokenUnverifiable(CanonicalException):
    code = 'BEARER_TOKEN_UNVERIFIABLE'
    http_status_code = 403
    message = (
        "The authenticity and integrity of the Bearer token could "
        "not be verified."
    )
    detail = (
        "The digital signature included in the Bearer token did "
        "not validate against any trusted key."
    )



"""Declares :class:`APIMetadataService`."""
from typing import List


class APIMetadataService:
    """Provides an interface to query the API metadata."""

    async def get(self, asgi, request):
        """Return a dictionary holding all API properties."""
        return {
            'capabilities': await self.capabilities(asgi, request),
            'catalog': await self.catalog(asgi, request)
        }

    async def catalog(self, asgi, request) -> dict:
        """Construct the API catalog."""
        return {
            'version': 'v1',
            'endpoints': {x.name: request.url_for(x.name) for x in asgi.routes}
        }

    async def capabilities(self, asgi, request) -> List[str]:
        """Return the list of API capabilities."""
        return []



import uvicorn
from frf import Resource
from frf import get_asgi_application


class AuthenticatedThing(Resource):
    authentication_classes = [None]
    required_scopes = ['thing.get']
    allow_cookie_auth = True

    async def retrieve(self, request, resource_id: int):
        return {

            "id": resource_id,
            "message": "Hello world!",
            'claims': request.state.claims.as_dict()
        }

    async def list(self):
        return [
            {"message": "First item in a list of AuthenticatedThing!"},
            {"message": "Second item in a list of AuthenticatedThing!"},
        ]


app = get_asgi_application(allowed_hosts=['*'])
app.add_resource(AuthenticatedThing)

if __name__ == '__main__':
    uvicorn.run('auth:app', host="127.0.0.1", port=8000, reload=True)



"""Declares routers for API endpoints."""
import inspect


class FastAPIRouter:
    """A router implementation for use with FastAPI."""
    __module__ = 'frf'

    def __init__(self):
        self.endpoints = {}

    def add_to_app(self, app):
        """Add all endpoints to the ASGI application."""
        for name, cls in dict.items(self.endpoints):
            base_path = f'/{name}'

            # Iterate over all members of the class and find actions.
            for name, func in inspect.getmembers(cls):
                if not hasattr(func, 'action'):
                    continue
                func.action.add_to_app(app, base_path, cls)

            if hasattr(cls, 'create'):
                cls.as_method_handler(
                    app, f'{base_path}', 'POST', 'create')
            if hasattr(cls, 'retrieve'):
                cls.as_method_handler(
                    app, f'{base_path}/{{pk}}', 'GET', 'retrieve')
            if hasattr(cls, 'list'):
                cls.as_method_handler(
                    app, f'{base_path}', 'GET', 'list')
            if hasattr(cls, 'update'):
                cls.as_method_handler(
                    app, f'{base_path}/{{pk}}', 'PATCH', 'update')
            if hasattr(cls, 'delete'):
                cls.as_method_handler(
                    app, f'{base_path}/{{pk}}', 'DELETE', 'delete')
            if hasattr(cls, 'replace'):
                cls.as_method_handler(
                    app, f'{base_path}/{{pk}}', 'PUT', 'replace')

            app.include_router(cls.router,
                tags=[cls.resource_name or cls.__name__])

    def register(self, name, cls):
        """Register an endpoint under the given name.

        Args:
            name (:class:`str`): the name of the endpoint and base path
                relative to the API root.
            cls (:class:`~frf.GenericViewSet`): the viewset or endpoint to mount
                at the given base path.

        Returns:
            None
        """
        self.endpoints[name] = cls



# pylint: skip-file
import asyncio

from unimatrix.ext.jose.tests.const import RSA_KEY
from unimatrix.ext import crypto
from unimatrix.ext import jose

import frf
from ..lib.test import ProtectedResourceTestCase
from .base import ProtectedResource


A = lambda *args, **kwargs: asyncio.run(*args, **kwargs)


class BearerTokenResolvedKeyValidationTestCase(ProtectedResourceTestCase):
    base_path = 'test'
    endpoints = [
        (base_path, ProtectedResource)
    ]

    def setUp(self):
        self.key = RSA_KEY
        crypto.chain.register(self.key.id, self.key)
        crypto.trust.register(self.key.id, A(self.key.get_public_key()))
        super().setUp()

    def get_signer(self):
        return crypto.get_signer(
            algorithm=crypto.algorithms.RSAPKCS1v15SHA256,
            keyid=self.key.id
        )

    def tearDown(self):
        crypto.chain.keys.pop(self.key.id)
        crypto.trust.keys.pop(self.key.id)

    def get_asgi_application(self):
        return frf.get_asgi_application(allowed_hosts=['*'])

    def test_no_authorization_header_returns_401(self):
        response = self.client.get('/test')
        self.assertEqual(response.status_code, 401)

        dto = response.json()
        self.assertIn('code', dto)
        self.assertEqual(dto['code'], 'BEARER_TOKEN_MISSING')

    def test_authorization_with_valid_token(self):
        response = self.client.get('/test', headers=self.headers)
        self.assertEqual(response.status_code, 200, msg=response.json())
